#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : Plus
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-11 11:51
   Description : description what the main function of this file
"""

from functools import reduce
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.exception.exception import StfEqualException, StfCondException


class Sum(Layer):
    def __init__(self, output_dim, fathers=None):
        fathers_output_dim = set(father.output_dim for father in fathers)
        if len(fathers_output_dim) > 1:
            raise StfCondException("len(fathers_output_dim)==1",
                                   "set(fathers_output_dim)={}".format(fathers_output_dim))
        elif fathers_output_dim != {output_dim}:
            raise StfEqualException("fathers_output_dim", "{output_dim}", fathers_output_dim, {output_dim})
        super(Sum, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w, x):
        return reduce(lambda a, b: a + b, x)

    def pull_back(self, w, x, y, ploss_py):
        ploss_pw = []
        ploss_px = dict(zip(self.fathers, [ploss_py] * len(x)))
        return ploss_pw, ploss_px


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : msb.py
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/2/15 上午11:27
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.pair import SharedPair, SharedTensorBase
from stensorflow.basic.basic_class.bitwise import PrivateTensorBitwise, SharedTensorBitwise
import tensorflow as tf
import numpy as np
import math
from stensorflow.exception.exception import StfDTypeException


def to_bool(x: tf.Tensor) -> tf.Tensor:
    """

    :param x:  tf.Tensor  shape =[s1,..., sn]
    :return:   tf.Tensor  shape =[s1,..., sn, sizeof(x.dtype)]
    """
    y = tf.bitcast(x, 'uint8')
    z = tf.bitwise.bitwise_and(tf.expand_dims(y, axis=-1), [1, 2, 4, 8, 16, 32, 64, 128])
    w = tf.bitwise.right_shift(z, [0, 1, 2, 3, 4, 5, 6, 7])
    new_shape = y.shape.as_list()
    new_shape[-1] = -1
    u = tf.reshape(w, new_shape)
    z = tf.cast(u, 'bool')
    return z


def bool_to_int64(z: tf.Tensor) -> tf.Tensor:
    """

    :param z: tf.Tensor of dtype bool  z.shape%64==0
    :return:  tf.Tensor of dtype int64 and shape = z.shape/64
    """
    u = tf.cast(z, 'uint8')
    u = tf.reshape(u, [-1, 8])
    u *= [1, 2, 4, 8, 16, 32, 64, 128]

    u = tf.reduce_sum(u, axis=[-1])

    u = tf.reshape(u, [-1, 8])
    u = tf.bitcast(u, 'int64')

    new_shape = z.shape.as_list()
    if new_shape[-1] % 64 != 0:
        raise Exception("must have new_shape[-1]%64 == 0")
    new_shape[-1] //= 64
    if new_shape[-1] == 1:
        new_shape = new_shape[0:-1]

    u = tf.reshape(u, new_shape)

    return u


def pad_to_64n(x: tf.Tensor):
    x = tf.reshape(x, [-1])
    if x.shape[0] % 64 == 0:
        return x
    else:
        n = math.ceil(x.shape[0] / 64)
        x = tf.pad(x, paddings=[[0, 64 * n - x.shape[0]]])
        return x


def bool_expansion(x: tf.Tensor):
    """

    :param x: tf.Tensor
    :return:  List of tf.Tensor of same type of x of shape [ceil([-1]/64)]
               bool expansion keeping the order
    """

    if x.dtype in [tf.int64, tf.uint64, tf.float64, tf.int32, tf.uint32, tf.float32, tf.int16, tf.uint16, tf.float16]:
        bit_size = x.dtype.size * 8
    else:
        raise StfDTypeException("x", "tf.intx,  tf.uintx, tf.floatx for x in 16,32,64", x.dtype)

    z = to_bool(x)
    list_z = tf.split(z, num_or_size_splits=bit_size, axis=-1)
    if x.dtype in [tf.float16, tf.float32, tf.float64]:
        for i in range(bit_size - 1):
            list_z[i] = tf.math.logical_xor(list_z[i], list_z[-1])
        list_z[-1] = tf.math.logical_not(list_z[-1])
    list_z = list(map(lambda zi: bool_to_int64(pad_to_64n(zi)), list_z))
    return list_z


def msb(x: SharedPair) -> SharedPair:
    with tf.device(x.ownerL):
        list_xL = bool_expansion(x.xL.inner_value)
        list_xL = list(
            map(lambda xi: PrivateTensorBitwise(owner=x.ownerL, inner_value=SharedTensorBitwise(inner_value=xi)),
                list_xL))

    with tf.device(x.ownerR):
        list_xR = bool_expansion(x.xR.inner_value)
        list_xR = list(
            map(lambda xi: PrivateTensorBitwise(owner=x.ownerR, inner_value=SharedTensorBitwise(inner_value=xi)),
                list_xR))

    carry = list_xL[0] * list_xR[0]
    for i in range(1, len(list_xL) - 1):
        carry = (list_xL[i] + carry) * (list_xR[i] + carry) - carry
    msb = list_xL[-1] + list_xR[-1] + carry
    with tf.device(x.ownerL):
        xL = to_bool(msb.xL.inner_value)
        xL = tf.cast(xL, 'int64')
        xL = SharedTensorBase(inner_value=xL, module=2)
    with tf.device(x.ownerR):
        xR = to_bool(msb.xR.inner_value)
        xR = tf.cast(xR, 'int64')
        xR = SharedTensorBase(inner_value=xR, module=2)
    y = SharedPair(ownerL=x.ownerL, ownerR=x.ownerR, xL=xL, xR=xR, fixedpoint=0)
    y = y.reshape([-1])
    if y.shape[0] != np.prod(x.shape):
        y = y[0: np.prod(x.shape)]
    return y.reshape(x.shape)


def special_mul(x, y):
    """
    :param x: =(x0, x1)
    :param y: =(y0, y1)
    :return: (x0y0, x1y2+x2)
    """
    return x[0] * y[0], x[0] * y[1] + x[1]


def reduce_special_mul(list_mat):
    """

    :param list_mat:
    :return:
    """
    lengh = len(list_mat)
    if lengh == 1:
        return list_mat[0]
    else:
        list0 = list_mat[0: lengh // 2]
        list1 = list_mat[lengh // 2: lengh]
        r0 = reduce_special_mul(list0)
        r1 = reduce_special_mul(list1)
        return special_mul(r0, r1)


def msb_log_round(x: SharedPair) -> SharedPair:
    with tf.device(x.ownerL):
        list_xL = bool_expansion(x.xL.inner_value)
        list_xL = list(
            map(lambda xi: PrivateTensorBitwise(owner=x.ownerL, inner_value=SharedTensorBitwise(inner_value=xi)),
                list_xL))

    with tf.device(x.ownerR):
        list_xR = bool_expansion(x.xR.inner_value)
        list_xR = list(
            map(lambda xi: PrivateTensorBitwise(owner=x.ownerR, inner_value=SharedTensorBitwise(inner_value=xi)),
                list_xR))

    matrix_list = [(xL + xR, xL * xR) for (xL, xR) in zip(list_xL, list_xR)]
    matrix_list.reverse()
    carry = reduce_special_mul(matrix_list[1:])[1]
    msb = list_xL[-1] + list_xR[-1] + carry
    with tf.device(x.ownerL):
        xL = to_bool(msb.xL.inner_value)
        xL = tf.cast(xL, 'int64')
        xL = SharedTensorBase(inner_value=xL, module=2)
    with tf.device(x.ownerR):
        xR = to_bool(msb.xR.inner_value)
        xR = tf.cast(xR, 'int64')
        xR = SharedTensorBase(inner_value=xR, module=2)
    y = SharedPair(ownerL=x.ownerL, ownerR=x.ownerR, xL=xL, xR=xR, fixedpoint=0)
    y = y.reshape([-1])
    if y.shape[0] != np.prod(x.shape):
        y = y[0: np.prod(x.shape)]
    return y.reshape(x.shape)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : ot
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-02 15:45
   Description : description what the main function of this file
"""
from stensorflow.global_var import StfConfig
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase
from stensorflow.basic.protocol.shift import cyclic_lshift


def assistant_ot(x: PrivateTensorBase, index: PrivateTensorBase, prf_flag=None,
                 compress_flag=None) -> SharedPairBase:
    if x.owner == index.owner:
        raise Exception("This function is for diffenrent owner")
    if x.shape[-1] != index.module:
        raise Exception("must have x.shape[-1]==index.module")
    if x.shape[:-1] != index.shape:
        raise Exception("must have x.shape[:-1]==index.shape")
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    z = cyclic_lshift(x=x, index=index, prf_flag=prf_flag, compress_flag=compress_flag)
    z = z[..., 0]
    return z


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group Copyright (c) 2021
   All Rights Reserved.
"""
from stensorflow.ml.nn.networks.NN import NN
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.relu import *
from stensorflow.ml.nn.layers.conv2d import Conv2dLocal
from stensorflow.ml.nn.layers.pooling import avg_pool2d, sum_pool2d_grad
from stensorflow.ml.nn.layers.flatten import *
from stensorflow.ml.nn.layers.dense import *
from stensorflow.ml.nn.layers.loss import *
from stensorflow.random import random


class LocalCNN(NN):
    """
    只有一层卷积核一层池化的CNN网络
    """
    def __init__(self, feature: PrivateTensor, label: Union[PrivateTensor, SharedPair],  loss=None):
        super(LocalCNN, self).__init__()
        # input layer, init data；
        # 这里将dim设置位输入的wight,后续不使用；仅仅是为了应用原有的模板
        layer = Input(dim=28, x=feature)
        local_layer_owner = layer.owner
        self.addLayer(ly=layer)
        # convolutional layer with 1 input channel, 16 output channels and a 5×5 filter
        layer = Conv2dLocal(output_dim=None, fathers=[layer], filters=16,
                            kernel_size=5, input_shape=[28, 28, 1], owner=local_layer_owner)
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU_Local(output_dim=layer.output_dim, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # Average pool
        layer = AveragePooling2DLocal(output_dim=None, fathers=[layer], pool_size=(2, 2), owner=local_layer_owner)
        self.addLayer(layer)
        # flatten data, only consider data_format = "NWHC"
        # 这里需要给出正确的output_dim，方便后续的全连接层
        layer = FlattenLocal(output_dim=None, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 全连接层
        # 这里添加一层，a 2304 × 100 linear layer
        # Dlayer = Dense_Local(output_dim=100, fathers=[layer], owner=local_layer_owner)
        # self.addLayer(Dlayer)
        # 添加一层Relu
        # Relu Layer
        # layer = ReLU_Local(output_dim=100, fathers=[Dlayer], owner=local_layer_owner)
        # self.addLayer(layer)
        # a 2304 × 10 linear layer； a 100* 10 line layer
        layer = Dense_Local(output_dim=10, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失计算
        layer_loss = CrossEntropyLossWithSoftmaxLocal(layer_score=layer, layer_label=layer_label, owner=local_layer_owner)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        self.cut_off()
        # 输入层
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # 输出层
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        record_num_ceil_mod_batch_size,
                        with_sigmoid):
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        # 分批写入文件
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def replace_weight(self, keras_weight):
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2dLocal):
                # 用传入的权重直接进行预测
                # kernel = PrivateTensor(owner=ly.owner)
                # kernel.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel
                # 用传入的权重进行训练
                kernel = ly.w[0]
                kernel.load_from_numpy(keras_weight[i])
                i += 1
            if isinstance(ly, Dense_Local):
                # 用传入的权重直接进行预测
                # kernel1 = PrivateTensor(owner=ly.owner)
                # kernel1.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel1
                # kernel2 = PrivateTensor(owner=ly.owner)
                # kernel2.load_from_numpy(keras_weight[i+1])
                # ly.w[1] = kernel2
                # 用传入的权重进行训练
                kernel1 = ly.w[0]
                kernel2 = ly.w[1]
                kernel1.load_from_numpy(keras_weight[i])
                kernel2.load_from_numpy(keras_weight[i+1])
                i += 2

    def save_model(self, sess, save_file_path, model_file_machine):
        res = []
        for ly in self.layers:
            if isinstance(ly, Dense_Local) or isinstance(ly, Conv2dLocal):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    weight = sess.run(weight)
                    res.append(weight)
        res = np.array(res)
        np.savez(save_file_path, weight=res)


class LocalNetworkB(NN):
    """
    两层卷积和两层池化的复杂网络
    """
    def __init__(self, feature: PrivateTensor, label: Union[PrivateTensor, SharedPair], loss=None):
        super(LocalNetworkB, self).__init__()
        # 这里将dim设置位输入的wight,后续不使用；仅仅是为了应用原有的模板
        layer = Input(dim=28, x=feature)
        local_layer_owner = layer.owner
        self.addLayer(layer)
        # convolutional layer with 1 input channel, 16 output channels and a 5×5 filter
        layer = Conv2dLocal(output_dim=None, fathers=[layer], filters=16,
                            kernel_size=5, input_shape=[28, 28, 1], owner=local_layer_owner)
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU_Local(output_dim=layer.output_dim, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # Average pool
        layer = AveragePooling2DLocal(output_dim=None, fathers=[layer],
                                      pool_size=(2, 2), owner=local_layer_owner)
        self.addLayer(layer)
        # 16 input channels, 16 output channels and another 5×5 filter
        layer = Conv2dLocal(output_dim=None, fathers=[layer], filters=16,
                            kernel_size=5, input_shape=layer.output_dim, owner=local_layer_owner)
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU_Local(output_dim=layer.output_dim, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # Average pool
        layer = AveragePooling2DLocal(output_dim=None, fathers=[layer], pool_size=(2, 2), owner=local_layer_owner)
        self.addLayer(layer)
        # flatten data, only consider data_format = "NWHC"
        # 这里需要给出正确的output_dim，方便后续的全连接层
        layer = FlattenLocal(output_dim=None, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 全连接层
        # 256×100 fully-connected layer
        layer = Dense_Local(output_dim=100, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # Relu Layer, 需要给出一个正确的output_dim
        layer = ReLU_Local(output_dim=100, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # a 100 × 10 linear layer
        layer = Dense_Local(output_dim=10, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失计算
        layer_loss = CrossEntropyLossWithSoftmaxLocal(layer_score=layer, layer_label=layer_label,
                                                      owner=local_layer_owner)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        self.cut_off()
        # 输入层
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # 输出层
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        record_num_ceil_mod_batch_size,
                        with_sigmoid):
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def save_model(self, sess, save_file_path, model_file_machine):
        res = []
        for ly in self.layers:
            if isinstance(ly, Dense_Local) or isinstance(ly, Conv2dLocal):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    weight = sess.run(weight)
                    res.append(weight)
        res = np.array(res)
        np.savez(save_file_path, weight=res)

    def replace_weight(self, keras_weight):
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2dLocal):
                # 用传入的权重直接进行预测
                # kernel = PrivateTensor(owner=ly.owner)
                # kernel.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel
                # 用传入的权重进行训练
                kernel = ly.w[0]
                kernel.load_from_numpy(keras_weight[i])
                i += 1
            if isinstance(ly, Dense_Local):
                # 用传入的权重直接进行预测
                # kernel1 = PrivateTensor(owner=ly.owner)
                # kernel1.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel1
                # kernel2 = PrivateTensor(owner=ly.owner)
                # kernel2.load_from_numpy(keras_weight[i+1])
                # ly.w[1] = kernel2
                # 用传入的权重进行训练
                kernel1 = ly.w[0]
                kernel2 = ly.w[1]
                kernel1.load_from_numpy(keras_weight[i])
                kernel2.load_from_numpy(keras_weight[i+1])
                i += 2


class LocalNetworkC(NN):
    """
    两层卷积和两层池化的复杂网络
    """
    def __init__(self, feature: PrivateTensor, label: Union[PrivateTensor, SharedPair], loss=None):
        super(LocalNetworkC, self).__init__()
        # 这里将dim设置位输入的wight,后续不使用；仅仅是为了应用原有的模板
        layer = Input(dim=28, x=feature)
        local_layer_owner = layer.owner
        self.addLayer(layer)
        # convolutional layer with 1 input channel, 16 output channels and a 5×5 filter
        layer = Conv2dLocal(output_dim=None, fathers=[layer], filters=20,
                            kernel_size=5, input_shape=[28, 28, 1], owner=local_layer_owner)
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU_Local(output_dim=layer.output_dim, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # Average pool
        layer = AveragePooling2DLocal(output_dim=None, fathers=[layer],
                                      pool_size=(2, 2), owner=local_layer_owner)
        self.addLayer(layer)


        # 20 input channels, 50 output channels and another 5×5 filter
        layer = Conv2dLocal(output_dim=None, fathers=[layer], filters=50,
                            kernel_size=5, input_shape=layer.output_dim, owner=local_layer_owner)
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU_Local(output_dim=layer.output_dim, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # Average pool
        layer = AveragePooling2DLocal(output_dim=None, fathers=[layer], pool_size=(2, 2), owner=local_layer_owner)
        self.addLayer(layer)
        # flatten data, only consider data_format = "NWHC"
        # 这里需要给出正确的output_dim，方便后续的全连接层
        layer = FlattenLocal(output_dim=None, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 全连接层
        # 800x500 fully-connected layer
        layer = Dense_Local(output_dim=500, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # a 500 × 10 linear layer
        layer = Dense_Local(output_dim=10, fathers=[layer], owner=local_layer_owner)
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失计算
        layer_loss = CrossEntropyLossWithSoftmaxLocal(layer_score=layer, layer_label=layer_label,
                                                      owner=local_layer_owner)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        self.cut_off()
        # 输入层
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # 输出层
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        with_sigmoid):
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def save_model(self, sess, save_file_path, model_file_machine):
        res = []
        for ly in self.layers:
            if isinstance(ly, Dense_Local) or isinstance(ly, Conv2dLocal):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    weight = sess.run(weight)
                    res.append(weight)
        res = np.array(res)
        np.savez(save_file_path, weight=res)

    def replace_weight(self, keras_weight):
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2dLocal):
                # 用传入的权重直接进行预测
                # kernel = PrivateTensor(owner=ly.owner)
                # kernel.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel
                # 用传入的权重进行训练
                kernel = ly.w[0]
                kernel.load_from_numpy(keras_weight[i])
                i += 1
            if isinstance(ly, Dense_Local):
                # 用传入的权重直接进行预测
                # kernel1 = PrivateTensor(owner=ly.owner)
                # kernel1.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel1
                # kernel2 = PrivateTensor(owner=ly.owner)
                # kernel2.load_from_numpy(keras_weight[i+1])
                # ly.w[1] = kernel2
                # 用传入的权重进行训练
                kernel1 = ly.w[0]
                kernel2 = ly.w[1]
                kernel1.load_from_numpy(keras_weight[i])
                kernel2.load_from_numpy(keras_weight[i+1])
                i += 2




#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : private
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-13 19:33
   Description : description what the main function of this file
"""

import tensorflow as tf
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase, get_device
from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.basic.operator.arithmetic import add, sub, mul, matmul, rmul, truediv
from stensorflow.basic.operator.order import greater, geq, less, leq
from stensorflow.global_var import StfConfig
from stensorflow.exception.exception import StfEqualWarning, StfCondException, StfEqualException, StfTypeException
import numpy as np


class PrivateTensor(PrivateTensorBase):
    def __init__(self, owner, fixedpoint: int = None, inner_value=None, module: int = None, op_map={}):
        super(PrivateTensor, self).__init__(owner, fixedpoint, inner_value, module)
        self.op_map = {"add": add, "sub": sub, "mul": mul, "matmul": matmul, "gt": greater, "ge": geq,
                       "lt": less, "le": leq, "rmul": rmul, "truediv": truediv}
        self.op_map.update(op_map)

    def __repr__(self) -> str:
        return 'PrivateTensor(owner={}, fixedpoint={}, module={}, shape={})'.format(self.owner.to_string(),
                                                                                    self.fixedpoint, self.module,
                                                                                    self.shape)

    @classmethod
    def from_PrivteTensorBase(cls, x: PrivateTensorBase, op_map={}):
        return PrivateTensor(owner=x.owner, inner_value=x.inner_value, fixedpoint=x.fixedpoint, module=x.module,
                             op_map=op_map)

    def transpose(self):
        with tf.device(self.owner):
            inner_value = tf.transpose(self.inner_value)
        return PrivateTensor(owner=self.owner, inner_value=inner_value, fixedpoint=self.fixedpoint, module=self.module,
                             op_map=self.op_map)

    def squeeze(self, axis):
        with tf.device(self.owner):
            inner_value = tf.squeeze(self.inner_value, axis=axis)
        return PrivateTensor(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value, module=self.module,
                             op_map=self.op_map)

    def expend_dims(self, axis):
        with tf.device(self.owner):
            inner_value = tf.expand_dims(self.inner_value, axis=axis)
        return PrivateTensor(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value, module=self.module,
                             op_map=self.op_map)

    def split(self, size_splits, axis=0, num=None):
        with tf.device(self.owner):
            inner_values = tf.split(self.inner_value, size_splits, axis, num)
            pvs = []
            for inner_value in inner_values:
                pvs += [
                    PrivateTensor(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value, module=self.module,
                                  op_map=self.op_map)]

        return tuple(pvs)

    def to_SharedPair(self, other_owner, op_map={}):
        """

        :param other_owner:
        :param op_map:
        :return:

        """
        xL = SharedTensorBase(inner_value=self.inner_value, module=self.module)
        xR = -xL.ones_like()  # use one but not zero, for div
        xL = xL + xL.ones_like()
        x = SharedPair(ownerL=self.owner, ownerR=other_owner, xL=xL, xR=xR, fixedpoint=self.fixedpoint, op_map=op_map)
        return x

    def zeros_like(self):
        with tf.device(self.owner):
            inner_value = tf.zeros_like(self.inner_value)
        x = PrivateTensor(owner=self.owner, inner_value=inner_value, module=self.module, fixedpoint=self.fixedpoint,
                          op_map=self.op_map)
        return x

    def ones_like(self, fixedpoint=None):
        if fixedpoint is None:
            fixedpoint = self.fixedpoint
        with tf.device(self.owner):
            inner_value = (1 << fixedpoint) * tf.ones_like(self.inner_value)
        x = PrivateTensor(owner=self.owner, inner_value=inner_value, module=self.module, fixedpoint=fixedpoint,
                          op_map=self.op_map)
        return x

    def down_bound_like(self):
        with tf.device(self.owner):
            inner_value = StfConfig.lower_bd_int64 * tf.ones_like(self.inner_value, dtype='int64')
        x = PrivateTensor(owner=self.owner, inner_value=inner_value, module=self.module, fixedpoint=self.fixedpoint,
                          op_map=self.op_map)
        return x

    def up_bound_like(self):
        with tf.device(self.owner):
            inner_value = StfConfig.upper_bd_int64 * tf.ones_like(self.inner_value, dtype='int64')
        x = PrivateTensor(owner=self.owner, inner_value=inner_value, module=self.module, fixedpoint=self.fixedpoint,
                          op_map=self.op_map)
        return x

    def concat(self, other, axis):
        if self.owner == other.owner:
            fixed_point = min(self.fixedpoint, other.fixedpoint)
            x = self.dup_with_precision(fixed_point)
            y = other.dup_with_precision(fixed_point)
            with tf.device(self.owner):
                inner_value = tf.concat([x.inner_value, y.inner_value], axis=axis)
            return PrivateTensor(self.owner, fixedpoint=fixed_point, inner_value=inner_value, module=self.module,
                                 op_map=self.op_map)
        else:
            x = self.to_SharedPair(other.owner)
            y = other.to_SharedPair(self.owner)
            return x.concat(y, axis=axis)

    def stack(self, other, axis):
        if self.owner == other.owner:
            fixed_point = min(self.fixedpoint, other.fixedpoint)
            x = self.dup_with_precision(fixed_point)
            y = other.dup_with_precision(fixed_point)
            with tf.device(self.owner):
                inner_value = tf.stack([x.inner_value, y.inner_value], axis=axis)
            return PrivateTensor(self.owner, fixedpoint=fixed_point, inner_value=inner_value, module=self.module,
                                 op_map=self.op_map)
        else:
            x = self.to_SharedPair(other.owner)
            y = other.to_SharedPair(self.owner)
            return x.stack(y, axis=axis)

    def to_compress_tensor(self, dtype: tf.dtypes.DType = tf.int64) -> tf.Tensor:
        with tf.device(self.owner):
            return self.to_SharedTensor().to_compress_tensor(dtype=dtype)

    def decompress_from(self, y: tf.Tensor, shape):
        with tf.device(self.owner):
            st = self.to_SharedTensor_like()
            st.decompress_from(y, shape=shape)
            self.inner_value = st.inner_value

    def dup_with_precision(self, new_fixedpoint: int):
        y = super(PrivateTensor, self).dup_with_precision(new_fixedpoint)
        return PrivateTensor.from_PrivteTensorBase(y, op_map=self.op_map)

    def reduce_sum(self, axis, keepdims=False):
        with tf.device(self.owner):
            inner_value = tf.reduce_sum(self.inner_value, axis=axis, keepdims=keepdims)
        return PrivateTensor(owner=self.owner, inner_value=inner_value,
                             fixedpoint=self.fixedpoint, module=self.module, op_map=self.op_map)

    def to_private(self, owner):
        owner = get_device(owner)
        if owner == self.owner:
            return self
        else:
            StfEqualWarning("owner", "self.owner", owner, self.owner)
            with tf.device(owner):
                inner_value = tf.identity(self.inner_value)
            return PrivateTensor(owner=owner, inner_value=inner_value,
                                 fixedpoint=self.fixedpoint, module=self.module, op_map=self.op_map)

    def load_from_numpy(self, x: np.ndarray):
        with tf.device(self.owner):
            self.load_from_tf_tensor(x)

    def _convert_result_type_(self, result):
        if isinstance(result, PrivateTensorBase):
            return self.from_PrivteTensorBase(x=result, op_map=self.op_map)
        elif isinstance(result, SharedPairBase):
            return SharedPair.from_SharedPairBase(result)
        else:
            raise StfTypeException("result", "PrivateTensorBase or SharedPairBase", type(result))

    def __add__(self, other):
        result = self.op_map['add'](self, other)
        return self._convert_result_type_(result)

    def __sub__(self, other):
        result = self.op_map['sub'](self, other)
        return self._convert_result_type_(result)

    def __mul__(self, other):
        result = self.op_map['mul'](self, other)
        return self._convert_result_type_(result)

    def __rmul__(self, other):
        result = self.op_map['rmul'](other, self)
        return self._convert_result_type_(result)

    def __truediv__(self, other):
        result = self.op_map['truediv'](self, other)
        return self._convert_result_type_(result)

    def __matmul__(self, other):
        result = self.op_map['matmul'](self, other)
        return self._convert_result_type_(result)

    def __pow__(self, power):
        if not isinstance(power, int):
            raise Exception("power must be a non-negative integral number.")
        elif power < 0:
            raise Exception("power must be a non-negative integral number.")
        elif power == 0:
            return self.ones_like()
        elif power == 1:
            return self
        elif power == 2:
            return (self * self).dup_with_precision(new_fixedpoint=self.fixedpoint)
        else:  # power > 2:
            power_helf = power // 2
            y_power_helf = self ** power_helf
            y = y_power_helf * y_power_helf
            if power % 2 == 0:
                return y.dup_with_precision(new_fixedpoint=self.fixedpoint)
            else:
                return (y * self).dup_with_precision(new_fixedpoint=self.fixedpoint)

    def __invert__(self):
        """
        :return:
        """
        if self.module is not None:
            raise StfCondException("self.module is None", "self.module={}".format(self.module))
        else:
            inner_value = (2 ** (2 * self.fixedpoint)) // self.inner_value
            return PrivateTensor(owner=self.owner, fixedpoint=self.fixedpoint,
                                 inner_value=inner_value, module=self.module)

    def __lt__(self, other):
        result = self.op_map['lt'](self, other)
        return self._convert_result_type_(result)

    def __le__(self, other):
        result = self.op_map['le'](self, other)
        return self._convert_result_type_(result)

    def __gt__(self, other):
        result = self.op_map['gt'](self, other)
        return self._convert_result_type_(result)

    def __ge__(self, other):
        result = self.op_map['ge'](self, other)
        return self._convert_result_type_(result)


class PrivateVariable(PrivateTensor):
    def __init__(self, owner, fixedpoint: int = 14, initial_inner_value=None, module: int = None, op_map={}):
        super(PrivateVariable, self).__init__(owner=owner, fixedpoint=fixedpoint, module=module, op_map=op_map)
        if initial_inner_value is not None:
            with tf.device(self.owner):
                if module is None:
                    self.inner_value = tf.Variable(initial_value=initial_inner_value, trainable=False)
                else:
                    self.inner_value = tf.Variable(initial_value=initial_inner_value % module, trainable=False)

    def __repr__(self) -> str:
        return 'PrivateVariable(owner={}, fixedpoint={}, module={}, shape={})'.format(self.owner.to_string(),
                                                                                      self.fixedpoint, self.module,
                                                                                      self.shape)

    def assign(self, other: PrivateTensorBase) -> tf.Operation:
        if other.owner != self.owner:
            raise StfEqualException("other.owner", "self.owner", other.owner, self.owner)
        if self.fixedpoint != other.fixedpoint:
            raise Exception("the fixedpoint must be same")
        if self.module != other.module:
            raise Exception("the fixedpoint must be same")
        with tf.device(self.owner):
            assign_op = self.inner_value.assign(value=other.inner_value, read_value=False)
        return assign_op

    def load_from_tf_tensor(self, x: tf.Tensor):
        """

        :param x: tf.Tensor
        :return:
        """
        with tf.device(self.owner):
            if self.module is not None:
                self.inner_value = tf.Variable(
                    initial_value=tf.cast(tf.multiply(x, (1 << self.fixedpoint)), 'int64') % self.module,
                    trainable=False)
            else:
                self.inner_value = tf.Variable(initial_value=tf.cast(tf.multiply(x, (1 << self.fixedpoint)), 'int64'),
                                               trainable=False)

    def load_from_numpy(self, x: np.ndarray, const=False):
        if const:
            super(PrivateVariable, self).load_from_tf_tensor(tf.constant(x))
        else:
            with tf.device(self.owner):
                self.load_from_tf_tensor(tf.constant(x))



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : fnz
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-04 19:13
   Description : Find the First Non-Zero bit:  https://eprint.iacr.org/2021/857

"""
import tensorflow as tf
from stensorflow.basic.basic_class.base import SharedTensorBase, SharedPairBase
from stensorflow.exception.exception import StfCondException
from stensorflow.global_var import StfConfig
from stensorflow.random.random import get_seed, gen_ZZn_from_seed, gen_ZZn
from sympy.ntheory.modular import isprime


class ElementInSemiDirectProductZZnFpxn:
    def __init__(self, p, n, shape):
        self.p = p
        self.n = n
        self.shape = shape
        self.i = SharedTensorBase(module=n, shape=shape)
        self.u = SharedTensorBase(module=p, shape=shape + [n])

    def from_SharedTensor(self, i: SharedTensorBase, u: SharedTensorBase):
        if i.module != self.n:
            raise Exception("must have i.module == self.n")
        if u.module != self.p:
            raise Exception("must have u.module == self.p")
        if i.shape != u.shape[:-1]:
            raise Exception("must have i.shape == u.shape[:-1]")
        self.i = i
        self.u = u  # LiTu

    def load_from_tf_tensor(self, i: tf.Tensor, u: tf.Tensor):
        if i.shape != u.shape[:-1]:
            raise Exception("must have i.shape == u.shape[:-1]")
        self.i.inner_value = tf.cast(i, 'int64') % self.n
        self.u.inner_value = tf.cast(u, 'int64') % self.p

    def from_uniform_distribution(self, u_shape, seed=None):
        i_shape = u_shape[:-1]
        if seed is None:
            # self.i = SharedTensorBase(inner_value=tf.random.uniform(shape=i_shape, minval=0, maxval=self.n,
            #                                                         dtype='int64', seed=seed), module=self.n)
            self.i = SharedTensorBase(inner_value=gen_ZZn(shape=i_shape, module=self.n), module=self.n)

            # self.u = SharedTensorBase(inner_value=tf.random.uniform(shape=u_shape, minval=1, maxval=self.p,
            #                                                         dtype='int64', seed=seed), module=self.p)
            self.u = SharedTensorBase(inner_value=1 + gen_ZZn(shape=u_shape, module=self.p - 1), module=self.p)
        else:
            self.i = SharedTensorBase(inner_value=gen_ZZn_from_seed(shape=i_shape, seed=seed, module=self.n),
                                      module=self.n)
            self.u = SharedTensorBase(inner_value=1 + gen_ZZn_from_seed(shape=u_shape, seed=seed, module=self.p - 1),
                                      module=self.p)

    # def from_uniform_PRF(self, seed, scope, id, u_shape):
    #     i_shape = u_shape[:-1]
    #     self.i = PRF_v0(seed=seed, scope=scope, id=id, shape=i_shape, module=self.n)
    #
    #     self.u = PRF_v0(seed=seed, scope=scope, id=id, shape=u_shape, module=self.p - 1)
    #
    #     self.u.inner_value = self.u.inner_value + 1
    #     self.u.module = self.p

    def __mul__(self, other):
        if not isinstance(other, ElementInSemiDirectProductZZnFpxn):
            raise Exception("other must be an Element_in_semi_direct_product_ZnZ_Fpxn")
        if self.n != other.n:
            raise Exception("must have self.n == other.n")
        if self.p != other.p:
            raise Exception("must have self.p == other.p")
        i = self.i + other.i
        u = self.u.rshift(other.i.inner_value) * other.u
        z = ElementInSemiDirectProductZZnFpxn(p=self.p, n=self.n, shape=self.shape)
        z.from_SharedTensor(i, u)
        # p.load_from_tf_tensor(i.inner_value, u.inner_value)
        return z

    def __invert__(self):  # L_iT_u L_{-i}T_(L_iu^{-1})=L_i T_u T_{u^-1} L_{-i}=1
        u = (self.u ** (self.p - 2)).lshift(self.i.inner_value)
        invert = ElementInSemiDirectProductZZnFpxn(p=self.p, n=self.n, shape=self.shape)
        invert.from_SharedTensor(-self.i, u)
        return invert

    # def __invert__(self):  # L_iT_u L_{-i}T_(L_iu^{-1})=L_i T_u T_{u^-1} L_{-i}=1
    #     return self

    def act(self, x: SharedTensorBase):
        if not isinstance(x, SharedTensorBase):
            raise Exception("other must be a SharedTensorBase")
        if self.n != x.shape[-1]:
            raise Exception("must have self.n == x.shape[-1]")
        if self.p != x.module:
            raise Exception("must have self.p == x.module")
        return (self.u * x).lshift(self.i.inner_value)

    def to_compress_tensor(self, dtype=tf.dtypes.int64):
        compressed_i = self.i.to_compress_tensor(dtype)
        compressed_u = self.u.to_compress_tensor(dtype)
        return (compressed_i, compressed_u)

    def decompress_from(self, compressed_i, compressed_u, shape_i, shape_u):
        self.i.decompress_from(compressed_i, shape_i)
        self.u.decompress_from(compressed_u, shape_u)


# def _G_module_recover_without_PRF(b: SharedPairBase):
#     """
#
#     :param b: in A
#     :return: (g, a) in G x A s.t ga=b
#     """
#     if b.xL.module is None:
#         raise Exception("must have x.xL.module is not None")
#     with tf.device(StfConfig.RS[0]):
#         g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
#         g.from_uniform_distribution(u_shape=b.xL.shape)
#
#         u = b.xR.random_uniform_adjoint()
#         v = g.act(u)
#         vL = v.random_uniform_adjoint()
#         vR = v - vL
#     with tf.device(b.ownerR):
#         c = b.xR - vR
#     with tf.device(b.ownerL):
#         w = (~g).act(b.xL - vL + c)
#     with tf.device(b.ownerR):
#         a = w + u
#     return (g, a)


# def _G_module_recover_without_compress(b: SharedPairBase, prf_flag=StfConfig.prf_flag):
#     """
#
#     :param b: in A
#     :return: (g, a) in G x A s.t ga=b
#     """
#     if b.xL.module is None:
#         raise Exception("must have x.xL.module is not None")
#     with tf.device(StfConfig.RS[0]):
#         g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
#         if prf_flag:
#             seed_g = get_seed()
#             seed_u = get_seed()
#             seed_v = get_seed()
#             g.from_uniform_distribution(u_shape=b.shape, seed=seed_g)
#             u = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_u)
#         else:
#             g.from_uniform_distribution(u_shape=b.xL.shape)
#             u = b.to_SharedTensor_like().random_uniform_adjoint()
#         v = g.act(u)
#         if prf_flag:
#             vL = v.random_uniform_adjoint(seed=seed_v)
#         else:
#             vL = v.random_uniform_adjoint()
#         vR = v - vL
#     with tf.device(b.ownerR):
#         c = b.xR - vR
#     with tf.device(b.ownerL):
#         if prf_flag:
#             g.from_uniform_distribution(u_shape=b.shape, seed=seed_g)
#             vL = v.random_uniform_adjoint(seed=seed_v)
#         w = (~g).act(b.xL - vL + c)
#     with tf.device(b.ownerR):
#         if prf_flag:
#             u = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_u)
#         a = w + u
#     return (g, a)
#
#
# def _G_module_recover_without_PRF_with_compress(b: SharedPairBase):
#     """
#
#     :param b: in A
#     :return: (g, a) in G x A s.t ga=b
#     """
#     if b.xL.module is None:
#         raise Exception("must have x.xL.module is not None")
#     with tf.device(StfConfig.RS[0]):
#         g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
#         g.from_uniform_distribution(u_shape=b.xL.shape)
#         u = b.xR.random_uniform_adjoint()
#         v = g.act(u)
#         vL = v.random_uniform_adjoint()
#         vR = v - vL
#         g_compress = g.to_compress_tensor()
#         u_compress = u.to_compress_tensor()
#         vL_compress = vL.to_compress_tensor()
#         vR_compress = vR.to_compress_tensor()
#     with tf.device(b.ownerR):
#         vR.decompress_from(vR_compress, vR.shape)
#         c = b.xR - vR
#         c_compress = c.to_compress_tensor()
#     with tf.device(b.ownerL):
#         g.decompress_from(g_compress[0], g_compress[1], g.i.shape, g.u.shape)
#         vL.decompress_from(vL_compress, vL.shape)
#         c.decompress_from(c_compress, c.shape)
#         w = (~g).act(b.xL - vL + c)
#         w_compress = w.to_compress_tensor()
#     with tf.device(b.ownerR):
#         w.decompress_from(w_compress, w.shape)
#         u.decompress_from(u_compress, u.shape)
#         a = w + u
#     return (g, a)


def _G_module_recover(b: SharedPairBase, prf_flag=None, compress_flag=None):
    """

    :param b: in A
    :return: (g, a) in G x A s.t ga=b
    """
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    if b.xL.module is None:
        raise Exception("must have x.xL.module is not None")
    with tf.device(StfConfig.RS[0]):
        g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
        if prf_flag:
            seed_g = get_seed()
            seed_u = get_seed()
            seed_v = get_seed()
            g.from_uniform_distribution(u_shape=b.shape, seed=seed_g)
            u = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_u)
        else:
            g.from_uniform_distribution(u_shape=b.shape)
            u = b.to_SharedTensor_like().random_uniform_adjoint()
        v = g.act(u)
        if prf_flag:
            vL = v.random_uniform_adjoint(seed=seed_v)
        else:
            vL = v.random_uniform_adjoint()
        vR = v - vL
        if compress_flag:
            if not prf_flag:
                g_compress = g.to_compress_tensor()
                u_compress = u.to_compress_tensor()
                vL_compress = vL.to_compress_tensor()
            vR_compress = vR.to_compress_tensor()
    with tf.device(b.ownerR):
        if compress_flag:
            vR.decompress_from(vR_compress, vR.shape)
        c = b.xR - vR
        if compress_flag:
            c_compress = c.to_compress_tensor()
    with tf.device(b.ownerL):
        if prf_flag:
            g.from_uniform_distribution(u_shape=b.shape, seed=seed_g)
            vL = v.random_uniform_adjoint(seed=seed_v)
        elif compress_flag:
            g.decompress_from(g_compress[0], g_compress[1], g.i.shape, g.u.shape)
            vL.decompress_from(vL_compress, vL.shape)
        if compress_flag:
            c.decompress_from(c_compress, c.shape)
        w = (~g).act(b.xL - vL + c)
        if compress_flag:
            w_compress = w.to_compress_tensor()
    with tf.device(b.ownerR):
        if prf_flag:
            u = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_u)
        elif compress_flag:
            u.decompress_from(u_compress, u.shape)
        if compress_flag:
            w.decompress_from(w_compress, w.shape)
        a = w + u
    return (g, a)


# def _G_module_recover(b: SharedPairBase, PRF: bool = False, compress=False):
#     if not PRF and not compress:
#         return _G_module_recover_without_PRF(b)
#     elif not PRF and compress:
#         return _G_module_recover_without_PRF_with_compress(b)
#     # elif PRF and compress:
#     #     return _G_module_recover_with_PRF_with_compress(b)
#     else:
#         raise NotImplementedError

#
# def _G_module_recover_inverse_without_PRF_without_compress(b: SharedPairBase):
#     """
#
#     :param b: in A
#     :return: (g, a) in G x A s.t (~g)a=b
#     """
#     if b.xL.module is None:
#         raise Exception("must have x.xL.module is not None")
#     with tf.device(StfConfig.RS[0]):
#         g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
#         g.from_uniform_distribution(u_shape=b.shape)
#         b_adjointR = b.xR.random_uniform_adjoint()
#         b_adjointL = b.xL.random_uniform_adjoint()
#         u = g.act(b_adjointL + b_adjointR)
#     with tf.device(b.ownerR):
#         delta_bR = b.xR - b_adjointR
#     with tf.device(b.ownerL):
#         delta_bL = b.xL - b_adjointL
#         w = g.act(delta_bL + delta_bR)
#     with tf.device(b.ownerR):
#         a = w + u
#     return g, a

#
# def _G_module_recover_inverse_without_PRF_with_compress(b: SharedPairBase):
#     """
#
#     :param b: in A
#     :return: (g, a) in G x A s.t (~g)a=b
#     """
#     if b.xL.module is None:
#         raise Exception("must have x.xL.module is not None")
#     with tf.device(StfConfig.RS[0]):
#         g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
#         g.from_uniform_distribution(u_shape=b.shape)
#         b_adjointR = b.xR.random_uniform_adjoint()
#         b_adjointL = b.xL.random_uniform_adjoint()
#         u = g.act(b_adjointL + b_adjointR)
#
#         g_compress_i, g_compress_u = g.to_compress_tensor()
#         u_compress = u.to_compress_tensor()
#         b_adjointL_compress = b_adjointL.to_compress_tensor()
#         b_adjointR_compress = b_adjointR.to_compress_tensor()
#     with tf.device(b.ownerR):
#         b_adjointR.decompress_from(b_adjointR_compress, b_adjointR.shape)
#         delta_bR = b.xR - b_adjointR
#         delta_bR_compress = delta_bR.to_compress_tensor()
#     with tf.device(b.ownerL):
#         g.decompress_from(shape_i=g.i.shape, shape_u=g.u.shape, compressed_i=g_compress_i, compressed_u=g_compress_u)
#         b_adjointL.decompress_from(b_adjointL_compress, b_adjointL.shape)
#         delta_bR.decompress_from(delta_bR_compress, delta_bR.shape)
#         delta_bL = b.xL - b_adjointL
#         w = g.act(delta_bL + delta_bR)
#         w_compress = w.to_compress_tensor()
#     with tf.device(b.ownerR):
#         w.decompress_from(w_compress, w.shape)
#         u.decompress_from(u_compress, u.shape)
#         a = w + u
#     return (g, a)


def _G_module_recover_inverse(b: SharedPairBase, prf_flag=None, compress_flag=None):
    """

    :param b: in A
    :return: (g, a) in G x A s.t (~g)a=b
    """
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    if b.xL.module is None:
        raise Exception("must have x.xL.module is not None")
    with tf.device(StfConfig.RS[0]):
        g = ElementInSemiDirectProductZZnFpxn(p=b.xL.module, n=b.shape[-1], shape=b.shape[0:-1])
        if prf_flag:
            seed_g = get_seed()
            seed_bL = get_seed()
            seed_bR = get_seed()
            g.from_uniform_distribution(u_shape=b.shape, seed=seed_g)
            b_adjointL = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_bL)
            b_adjointR = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_bR)
        else:
            g.from_uniform_distribution(u_shape=b.shape)
            b_adjointR = b.to_SharedTensor_like().random_uniform_adjoint()
            b_adjointL = b.to_SharedTensor_like().random_uniform_adjoint()

        u = g.act(b_adjointL + b_adjointR)
        if compress_flag:
            if not prf_flag:
                g_compress_i, g_compress_u = g.to_compress_tensor()
                b_adjointL_compress = b_adjointL.to_compress_tensor()
                b_adjointR_compress = b_adjointR.to_compress_tensor()
            u_compress = u.to_compress_tensor()

    with tf.device(b.ownerR):
        if prf_flag:
            b_adjointR = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_bR)
        elif compress_flag:
            b_adjointR.decompress_from(b_adjointR_compress, b_adjointR.shape)
        delta_bR = b.xR - b_adjointR
        if compress_flag:
            delta_bR_compress = delta_bR.to_compress_tensor()
    with tf.device(b.ownerL):
        if prf_flag:
            g.from_uniform_distribution(u_shape=b.shape, seed=seed_g)
            b_adjointL = b.to_SharedTensor_like().random_uniform_adjoint(seed=seed_bL)
        elif compress_flag:
            g.decompress_from(shape_i=g.i.shape, shape_u=g.u.shape, compressed_i=g_compress_i,
                              compressed_u=g_compress_u)
            b_adjointL.decompress_from(b_adjointL_compress, b_adjointL.shape)
        if compress_flag:
            delta_bR.decompress_from(delta_bR_compress, delta_bR.shape)
        delta_bL = b.xL - b_adjointL
        w = g.act(delta_bL + delta_bR)
        if compress_flag:
            w_compress = w.to_compress_tensor()
    with tf.device(b.ownerR):
        if compress_flag:
            w.decompress_from(w_compress, w.shape)
            u.decompress_from(u_compress, u.shape)
        a = w + u
    return (g, a)


def fnz_v1(x: SharedPairBase, prf_flag=None, compress_flag=None):
    """

    :param b: SharedPair with module is a prime number p and 0-1 value,
    there exists at least one bit i such that x[i]==1
    and p>=b.shape[-1]+2
    :return: SharedPair, the first non-zero bit along axis=-1
    """
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    p = x.xL.module
    if not isprime(p):
        raise StfCondException("p be a prime number", "p={}".format(p))
    if p < x.shape[-1] + 2:
        raise Exception("must have p>=x.shape[-1]+2")

    x_p = x.cumulative_sum(axis=-1)
    y = x_p - 2 * x + x.ones_like()

    g, a = _G_module_recover(y, prf_flag=prf_flag, compress_flag=compress_flag)

    with tf.device(x.ownerL):
        i = g.i
    with tf.device(x.ownerR):
        equel0_onehot = tf.cast(tf.equal(a.inner_value, 0), 'int32')

        _range = range(equel0_onehot.shape.as_list()[-1])
        j = tf.tensordot(equel0_onehot, _range, axes=[-1, 0])
        j = tf.cast(j, 'int64')
        j = SharedTensorBase(inner_value=j, module=len(_range))
    return SharedPairBase(ownerL=x.ownerL, ownerR=x.ownerR, xL=-i, xR=j, fixedpoint=0)


def fnz_v2(x: SharedPairBase, prf_flag=None, compress_flag=None):
    """

    :param b: SharedPair with module is a prime number p and 0-1 value,
    there exists at least one bit i such that x[i]==1
    and p>=b.shape[-1]+2
    :return: SharedPair, the first non-zero bit along axis=-1
    """
    p = x.xL.module
    if not isprime(p):
        raise Exception("x.xL.module must be an integer")
    if p < x.shape[-1] + 2:
        raise Exception("must have p>=x.shape[-1]+2")
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    x_p = x.cumulative_sum(axis=-1)
    y = x_p - 2 * x + x.ones_like()

    g, a = _G_module_recover(y, prf_flag=prf_flag, compress_flag=compress_flag)

    with tf.device(x.ownerL):
        i = g.i
    with tf.device(x.ownerR):
        _range = range(a.shape[-1])
        j = tf.reduce_sum(tf.where(tf.equal(a.inner_value, 0), _range, [0]), axis=-1)
        j = tf.cast(j, 'int64')
        j = SharedTensorBase(inner_value=j, module=len(_range))
    return SharedPairBase(ownerL=x.ownerL, ownerR=x.ownerR, xL=-i, xR=j, fixedpoint=0)


def fnz_v3(x: SharedPairBase, prf_flag=None, compress_flag=None):
    """
    利用 _G_module_recover_inverse 进行 fnz
    :param b: SharedPair with module is a prime number p and 0-1 value,
    there exists at least one bit i such that x[i]==1
    and p>=b.shape[-1]+2
    :return: SharedPair, the first non-zero bit along axis=-1
    """
    p = x.xL.module
    if not isprime(p):
        raise Exception("x.xL.module must be an integer")
    if p < x.shape[-1] + 2:
        raise Exception("must have p>=x.shape[-1]+2")
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    x_p = x.cumulative_sum(axis=-1)
    y = x_p - 2 * x + x.ones_like()
    g, a = _G_module_recover_inverse(y, prf_flag=prf_flag, compress_flag=compress_flag)  # gy=a
    with tf.device(x.ownerL):
        i = g.i
    with tf.device(x.ownerR):
        j = tf.argmin(input=a.inner_value, axis=-1)
        j = tf.cast(j, 'int64')
        j = SharedTensorBase(inner_value=j, module=y.shape[-1])
    return SharedPairBase(ownerL=x.ownerL, ownerR=x.ownerR, xL=i, xR=j, fixedpoint=0)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : shift
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-12-01 20:09
   Description : description what the main function of this file
"""
from stensorflow.basic.basic_class.base import SharedPairBase
from stensorflow.basic.basic_class.base import PrivateTensorBase
from stensorflow.random.random import get_seed
import tensorflow as tf
from stensorflow.global_var import StfConfig


def cyclic_lshift(index: PrivateTensorBase, x: PrivateTensorBase, prf_flag=None, compress_flag=None) -> SharedPairBase:
    if x.owner == index.owner:
        raise Exception("This function is for diffenrent owner")
    if x.shape[-1] != index.module:
        raise Exception("must have x.shape[-1]==index.module")
    if x.shape[:-1] != index.shape:
        raise Exception("must have x.shape[:-1]==index.shape")
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    with tf.device(StfConfig.RS[0]):
        # STEP 1 :  Random Server generate index_adjoint and x_adjoint, y_L, y_R, s.t
        #  L_{index_adjoint} x_adjoint = y_L + y_R
        if prf_flag:
            seed_index = get_seed()
            seed_x = get_seed()
            seed_y = get_seed()
        else:
            seed_index = None
            seed_x = None
            seed_y = None

        index_adjoint = index.to_SharedTensor_like().random_uniform_adjoint(seed_index)
        x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_x)
        y = x_adjoint.lshift(index_adjoint.inner_value)
        yL = y.random_uniform_adjoint(seed_y)
        yR = y - yL

        if compress_flag:
            if not prf_flag:
                x_adjoint_compress = x_adjoint.to_compress_tensor()
                index_adjoint_compress = index_adjoint.to_compress_tensor()
                yL_compress = yL.to_compress_tensor()
            yR_compress = yR.to_compress_tensor()

    # STEP 2: x.owner compute delta_x=x-x_adjoint
    with tf.device(x.owner):
        if prf_flag:
            x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_x)
        elif compress_flag:
            x_adjoint.decompress_from(x_adjoint_compress, shape=x_adjoint.shape)
        delta_x = x.to_SharedTensor() - x_adjoint
        if compress_flag:
            delta_x_compress = delta_x.to_compress_tensor()

    # Step 3. index.owner compute delta_i = index - index_adjoint
    #         and zL:=L_{index}(delta_x)+L_{delta_i}(y_L)
    with tf.device(index.owner):
        if prf_flag:
            index_adjoint = index.to_SharedTensor_like().random_uniform_adjoint(seed_index)
            yL = y.random_uniform_adjoint(seed_y)
        elif compress_flag:
            index_adjoint.decompress_from(index_adjoint_compress, shape=index_adjoint.shape)
            yL.decompress_from(yL_compress, yL.shape)
        delta_i = index.to_SharedTensor() - index_adjoint

        if compress_flag:
            delta_x.decompress_from(delta_x_compress, shape=delta_x.shape)
            delta_i_compress = delta_i.to_compress_tensor()

        zL = delta_x.lshift(index.inner_value) + yL.lshift(delta_i.inner_value)

    # Step 4.  x.owner compute zR:=L_{delta_i}(yR)
    with tf.device(x.owner):
        if compress_flag:
            delta_i.decompress_from(delta_i_compress, delta_i.shape)
            yR.decompress_from(yR_compress, yR.shape)
        zR = yR.lshift(delta_i.inner_value)
    z = SharedPairBase(ownerL=index.owner, ownerR=x.owner, xL=zL, xR=zR, fixedpoint=x.fixedpoint)
    return z



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : __init__.py
   Author : qizhi.zqz
   Email: qizhi.zqz@alibaba-inc.com
   Create Time : 2020-09-11 11:49
   Description : description what the main function of this file
"""


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : NETWORKC
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/9/2 上午11:02
   Description : description what the main function of this file
"""


from stensorflow.ml.nn.networks.NN import NN
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.relu import ReLU
from stensorflow.ml.nn.layers.pooling import AveragePooling2D, MaxPooling2D
from stensorflow.ml.nn.layers.conv2d import Conv2d
from stensorflow.basic.basic_class.pair import SharedVariablePair
from stensorflow.ml.nn.layers.loss import CrossEntropyLossWithSoftmax
from stensorflow.ml.nn.layers.flatten import Flatten
from stensorflow.ml.nn.layers.dense import Dense
from stensorflow.random import random
import numpy as np


class NetworkD(NN):
    """4 layer CNN.  NetworkC in SecureNN """
    def __init__(self, feature, label):
        super(NetworkD, self).__init__()
        # input layer, init data；
        layer = Input(dim=28, x=feature)
        self.addLayer(layer)
        # convolutional layer with 1 input channel, 5 output channels and a 5×5 filter, stride=2
        layer = Conv2d(output_dim=None, fathers=[layer], filters=5, strides=[1,2,2,1],
                       kernel_size=5, input_shape=[28, 28, 1])
        self.addLayer(layer)
        print("output_dim of Conv2D=", layer.output_dim)
        # Relu Layer
        layer = ReLU(output_dim=layer.output_dim, fathers=[layer])
        self.addLayer(layer)


        # flatten data, only consider data_format = "NWHC"
        layer = Flatten(output_dim=None, fathers=[layer])
        self.addLayer(layer)

        # 全连接层
        # 800×500 fully-connected layer
        layer = Dense(output_dim=100, fathers=[layer])
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU(output_dim=layer.output_dim, fathers=[layer])
        self.addLayer(layer)
        # a 500 × 10 linear layer
        layer = Dense(output_dim=10, fathers=[layer])
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失层
        layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        """
        Generates output predictions for the input samples.
        :param x: Input samples.
        :param out_prob: True or False
        :return: The predicted result or probability
        """
        self.cut_off()
        # Input Layer
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # Output Layer
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        with_sigmoid):
        """
        Save prediction results to file
        Computation is done in batches.
        :param sess: session
        :param x: PrivateTensor. Input samples.
        :param predict_file_name: String.
        :param pred_batch_num: Number of samples per batch.
        :param model_file_machine: String.
        :param record_num_ceil_mod_batch_size:
        :param with_sigmoid:
        :return:
        """
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        # 分批写入文件
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def print(self, model_file_machine, sess):
        """
        print model weights
        """
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Conv2d):
                print(ly)
                for weight in ly.w:
                    print(weight)
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    print("***")
                    print(sess.run(weight))

    def replace_weight(self, keras_weight):
        """
        将传入的权重赋值给当前网络
        """
        i = 0
        for ly in self.layers:
            if isinstance(ly, Conv2d):
                # 用传入的权重直接进行预测
                # kernel = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                # kernel.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel
                # 用传入的权重进行训练
                kernel = ly.w[0]
                assert isinstance(kernel, SharedVariablePair)
                kernel.load_from_tf_tensor(keras_weight[i])
                i += 1

            if isinstance(ly, Dense):
                # 用传入的权重直接进行预测
                # kernel1 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                # kernel1.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel1
                # kernel2 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i+1].shape)
                # kernel2.load_from_numpy(keras_weight[i+1])
                # ly.w[1] = kernel2
                # 用传入的权重进行训练
                kernel1 = ly.w[0]
                assert isinstance(kernel1, SharedVariablePair)
                kernel1.load_from_tf_tensor(keras_weight[i])
                kernel2 = ly.w[1]
                assert isinstance(kernel2, SharedVariablePair)
                kernel2.load_from_tf_tensor(keras_weight[i+1])
                i += 2

    def save_model(self, sess, save_file_path, model_file_machine):
        """
        Save the model weights to npz file.
        :param sess: session
        :param save_file_path: String, path to save the model.
        :param model_file_machine: String, the machine name.
        :return:
        """
        res = []
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Conv2d):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    weight = sess.run(weight)
                    res.append(weight)
        res = np.array(res)
        np.savez(save_file_path, weight=res)




#!/usr/bin/env python
# coding=utf-8

from stensorflow.ml.nn.networks.NN import NN
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.relu import ReLU
from stensorflow.basic.basic_class.pair import SharedVariablePair
from stensorflow.ml.nn.layers.loss import CrossEntropyLossWithSoftmax
from stensorflow.ml.nn.layers.dense import Dense
from stensorflow.random import random



class NETWORKA(NN):
    def __init__(self, feature, label):
        super(NETWORKA, self).__init__()
        # input layer, init data；
        layer = Input(dim=28*28, x=feature)
        self.addLayer(layer)
        # 全连接层
        # a 28*28 × 100 linear layer
        layer = Dense(output_dim=128, fathers=[layer])
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU(output_dim=128, fathers=[layer])
        self.addLayer(layer)
        # 全连接层
        # a 28*28 × 128 linear layer
        layer = Dense(output_dim=128, fathers=[layer])
        self.addLayer(layer)
        # Relu Layer
        layer = ReLU(output_dim=128, fathers=[layer])
        self.addLayer(layer)
        # 100x10
        layer = Dense(output_dim=10, fathers=[layer])
        self.addLayer(layer)
        # 输出层
        layer_label = Input(dim=10, x=label)
        self.addLayer(ly=layer_label)
        # 损失计算
        layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
        self.addLayer(ly=layer_loss)

    def predict(self, x, out_prob=True):
        self.cut_off()
        # 输入层
        l_input = self.layers[0]
        assert isinstance(l_input, Input)
        l_input.replace(x)
        self.layers[0] = l_input
        # 输出层
        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()
        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        pred_batch_num, model_file_machine,
                        with_sigmoid):
        y_pred = self.predict(x=x,  out_prob=with_sigmoid)
        id_y_pred = y_pred.to_tf_str(owner=model_file_machine)
        random.random_init(sess)
        # with open(predict_file_name, "w") as f:
        #     for batch in range(batch_num - 1):
        #         records = sess.run(id_y_pred)
        #         records = "\n".join(records.astype('str'))
        #         f.write(records + "\n")
        #
        #     records = sess.run(id_y_pred)[0:record_num_ceil_mod_batch_size]
        #     records = "\n".join(records.astype('str'))
        #     f.write(records + "\n")
        # 分批写入文件
        with open(predict_file_name, "w") as f:
            for batch in range(pred_batch_num):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                # records.to_file()
                f.write(records + "\n")

    def replace_weight(self, keras_weight):
        i = 0
        for ly in self.layers:
            if isinstance(ly, Dense):
                # kernel1 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i].shape)
                # kernel1.load_from_numpy(keras_weight[i])
                # ly.w[0] = kernel1
                # kernel2 = SharedPair(ownerL="L", ownerR="R", shape=keras_weight[i+1].shape)
                # kernel2.load_from_numpy(keras_weight[i+1])
                # ly.w[1] = kernel2
                # 分割
                kernel1 = ly.w[0]
                assert isinstance(kernel1, SharedVariablePair)
                kernel1.load_from_tf_tensor(keras_weight[i])
                kernel2 = ly.w[1]
                assert isinstance(kernel2, SharedVariablePair)
                kernel2.load_from_tf_tensor(keras_weight[i+1])
                i += 2





#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : logistic_regression
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 11:42
   Description : description what the main function of this file
"""
import tensorflow as tf
import stensorflow as stf
from typing import Union
from stensorflow.basic.basic_class.pair import SharedPair, SharedVariablePair
from stensorflow.exception.exception import StfValueException
from stensorflow.random.random import random_init
from stensorflow.homo_enc.homo_enc import homo_init
from stensorflow.basic.basic_class.base import get_device
from stensorflow.basic.basic_class.private import PrivateTensor
import numpy as np
from stensorflow.global_var import StfConfig
import os
from stensorflow.basic.operator.sigmoid import sigmoid_sin, sigmoid_idea


class LogisticRegression:
    """Logistic regression that features in 1 player."""

    def __init__(self, num_features, learning_rate=0.01, l2_regularization=1E-2):
        """

        :param num_features:  the number of features
        :param learning_rate:  learning rate
        :param l2_regularization: l2 regularization
        """

        self.num_features = num_features
        self.w = SharedVariablePair(ownerL="L", ownerR="R", shape=[num_features, 1])
        self.w.load_from_numpy(np.random.uniform(size=[num_features, 1],
                                                 low=-1.0 / np.sqrt(num_features + 1),
                                                 high=1.0 / np.sqrt(num_features + 1)))
        self.b = SharedVariablePair(ownerL="L", ownerR="R", shape=[1])
        self.b.load_from_tf_tensor(tf.zeros([1]))

        self.learning_rate = learning_rate
        self.l2_regularzation = l2_regularization

    @property
    def weights(self):
        return self.w, self.b

    def forward(self, x, with_sigmoid=True):
        """

        :param x: list<Union[PrivateTensor, SharedPair]>
        :param with_sigmoid:
        :return:
        """

        m = x @ self.w
        print("m.xL.shape=", m.xL.shape)
        print("m.xR.shape=", m.xR.shape)
        out = m + self.b

        if with_sigmoid:
            y = sigmoid_sin(out, M=16)
            # y = sigmoid_idea(out, M=16)
        else:
            y = out
        return y

    def backward(self, x: Union[PrivateTensor, SharedPair], dy: Union[PrivateTensor, SharedPair], learning_rate=0.01):
        """
        :param x:    fetures
        :param dy:    ploss/py
        :param learning_rate:   learning rate
        :return:
        """

        batch_size = x.shape[0]
        dw = x.transpose() @ (dy / batch_size) + self.l2_regularzation * self.w
        db = dy.reduce_sum(axis=0) / batch_size

        assign_ops = [
            self.w.assign(self.w - learning_rate * dw),
            self.b.assign(self.b - learning_rate * db),
        ]
        return assign_ops

    def loss_grad(self, y: Union[PrivateTensor, SharedPair], y_hat: Union[PrivateTensor, SharedPair]):
        """

        :param y:
        :param y_hat:
        :return:
        """
        dy = y_hat - y
        return dy

    def fit_batch(self, x: Union[PrivateTensor, SharedPair], y: Union[PrivateTensor, SharedPair]):
        """

        :param x:
        :param y:
        :return:
        """
        y_hat = self.forward(x)
        dy = self.loss_grad(y, y_hat)
        fit_batch_op = self.backward(x, dy, self.learning_rate)
        return fit_batch_op

    def fit(self, sess, x: Union[PrivateTensor, SharedPair], y: Union[PrivateTensor, SharedPair], num_batches,
            progress_file="./train_progress"):
        """

        :param sess:
        :param x:
        :param y:
        :param num_batches:
        :param progress_file:
        """
        fit_batch_op = self.fit_batch(x, y)
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            pre_produce_op = tf.group(StfConfig.pre_produce_list)
        if sess is not None:
            sess.run(tf.compat.v1.global_variables_initializer())
            with open(progress_file, "a") as f:

                for batch in range(num_batches):
                    print("batch=", batch)
                    if StfConfig.pre_produce_flag and StfConfig.offline_model:
                        sess.run(pre_produce_op)
                    else:
                        sess.run(fit_batch_op)
                    if batch % (int(num_batches / 100) + 1) == 0:
                        f.write(str(1.0 * batch / num_batches) + "\n")
                        f.flush()

    def predict_batch(self, x: Union[PrivateTensor, SharedPair], owner):
        """

        :param x:
        :return:
        """
        with tf.name_scope("predict_batch"):
            y_hat = self.forward(x, with_sigmoid=False)

            y_hat = y_hat.to_tf_tensor(owner=owner)

            y_hat = tf.sigmoid(y_hat)
            return y_hat

    # def predict(self, sess, x, file_name, num_batches, idx, owner, record_num_ceil_mod_batch_size,
    #             y: PrivateTensor = None, new_write_file=True):
    #     """
    #
    #     :param sess:
    #     :param x:
    #     :param file_name:
    #     :param num_batches:
    #     :param idx:
    #     :param owner: output owner
    #     :param record_num_ceil_mod_batch_size:
    #     """
    #     owner = get_device(owner)
    #     predict_batch = self.predict_batch(x, owner)
    #
    #     with tf.device(owner):
    #         predict_batch = tf.strings.as_string(predict_batch)
    #         if y is None:
    #
    #             predict_batch = tf.concat([idx, predict_batch], axis=1)
    #         else:
    #             y = tf.strings.as_string(y.to_tf_tensor())
    #             predict_batch = tf.concat([idx, predict_batch, y], axis=1)
    #         predict_batch = tf.compat.v1.reduce_join(predict_batch, axis=1, separator=", ")
    #
    #         if new_write_file:
    #             open_method = "w"
    #         else:
    #             open_method = "a"
    #         if sess is not None:
    #             with open(file_name, open_method) as f:
    #                 for batch in range(num_batches):
    #                     records = sess.run(predict_batch)
    #                     if batch == num_batches - 1:
    #                         records = records[0:record_num_ceil_mod_batch_size]
    #                     records = "\n".join(records.astype('str'))
    #                     f.write(records + "\n")

    def predict(self, id, x_test, pred_batch_num, sess, predict_file=None):
        if predict_file is None:
            predict_file = StfConfig.predict_to_file
        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            StfConfig.pre_produce_list = []
        y_pred = self.predict_batch(x_test, owner="R")

        if StfConfig.pre_produce_flag and StfConfig.offline_model:
            pre_produce_op = tf.group(StfConfig.pre_produce_list)
            homo_init(sess)
        random_init(sess)  # initialize the random model using in predict

        with open(predict_file, "w") as f:
            for batch in range(pred_batch_num):
                if StfConfig.pre_produce_flag and StfConfig.offline_model:
                    sess.run(pre_produce_op)
                else:
                    ids, records = sess.run([id, y_pred])
                    ids = ids.astype('str')
                    records = records.astype('str')
                    records = np.concatenate([ids, records], axis=1)
                    records = "\n".join([",".join(line) for line in records])
                    f.write(records + "\n")

    def save(self, model_file_path, model_file_machine="R"):

        if not os.path.exists(model_file_path):
            raise Exception("path not exist.{}".format(model_file_path))
        save_ops = []
        for i in range(len(self.weights)):
            model_file_path_i = os.path.join(model_file_path, "param_{i}".format(i=i))
            save_op = self.weights[i].reshape([-1]).to_private(model_file_machine).to_file(
                path=model_file_path_i, separator=", ", precision=StfConfig.to_str_precision)
            save_ops = save_ops + [save_op]
        save_op = tf.group(*save_ops)
        return save_op

    def save_as_plaintext(self, model_file_path, model_file_machine="R"):
        weights = self.b.to_private(model_file_machine).concat(self.w.reshape([-1]).to_private(model_file_machine),
                                                               axis=0)

        save_op = weights.to_file(path=model_file_path, separator=", ", precision=StfConfig.to_str_precision)
        return save_op

    def load(self, model_file_path, model_file_machine="R"):

        self.w = stf.PrivateVariable(owner=model_file_machine)
        self.b = stf.PrivateVariable(owner=model_file_machine)

        self.w.load_first_line_from_file(os.path.join(model_file_path, "param_0"), col_num=self.num_features)
        self.w = self.w.reshape([self.num_features, 1])

        self.b.load_first_line_from_file(os.path.join(model_file_path, "param_1"), col_num=1)
        self.b = self.b.reshape([1, 1])
